/*
 * innova interface for netlink I/O and raw I/O for
 * reinjection of packets, interface for portable 
 * shell script runned for get ULOG target over 
 * iptables for select group mask free.
 *
 * library for manage netlink packet is taken from
 * ulog, you should find more information about 
 * him author, laforge, on the header of him file,
 * reported on libipulogd/, and on him site about 
 * entry project:
 *
 * http://www.gnumonks.org/projects/ulogd
 *
 * this interface are used for manage reding of 
 * packets from netlink device. about netlink 
 * device you should read the man page netlink(7),
 *
 * in this file is impelemented a lot of function,
 * the four exported are:
 *
 * netlink_init 
 * get_packets
 * reinject_packet
 * cleanup_netlink
 *
 * ... is too easy, the code following is self
 * explicative :) you should read them and libipulog
 * for undestound better how this network hack
 * (steal packets from userspace keeping the kernel to
 * tracking and managing the session)
 *
 * 0.0.1:		datalink/netlink retrivering and
 * 			packet analysis, I/O packet handling
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <asm/types.h>
#include <net/if.h>

#include <linux/if_packet.h>
#include <linux/if_ether.h>
#include <linux/netlink.h>

#include "innova.h"
#include "libipulog.h"

extern int errno;
extern int ipulog_errno;
extern struct innova_struct *current;

static char ipt_cmd_str_local_mark[MAXCOMMANDLEN];
static char ipt_cmd_str_local_ulog[MAXCOMMANDLEN];
static char ipt_cmd_str_local_masq[MAXCOMMANDLEN];

static char iproute_cmd_local_dup[MAXCOMMANDLEN];

static char ipt_cmd_str_remote[MAXCOMMANDLEN];

void fill_portopt(int p, unsigned char direction, char *str, int len, struct servent *rem, struct servent *loc)
{
	char srcprefix, dstprefix;

	if(direction ==REMOTE)
	{
		srcprefix ='s';
		dstprefix ='d';
	}
	else
	{
		srcprefix ='d';
		dstprefix ='s';
	}

	memset(str, 0x00, len);
	
	/* actually is not supported any other kind of protocol */
	if(p ==IPPROTO_TCP || p ==IPPROTO_UDP)
	{
		if(rem !=NULL)
			snprintf(str, len, " -p %d --%cport %d", p, srcprefix, ntohs(rem->s_port));
		if(loc !=NULL)
			snprintf(str, len, "%s -p %d --%cport %d", str, p, dstprefix, ntohs(loc->s_port));
	}
	else
		memset(str, 0x00, len);
}

/*
 * this api is very example of lazy programming,
 * I needed to undestound witch netlink group (options for -j ULOG extension)
 * and witch mark (options for -j MARK) are free or what are been used,
 * for this reason I read the output generated by a piped manipoulated
 * output generated with iptables -L etc ..
 * on this weird programmed function, I'm supposing 
 * the presence of grep and cut and sed on evry systems.
 *
 * NETLINK_GROUP, MARK, IP_TABLE is simply define for undestound what 
 * command should be used for track 
 */
#define NETLINK_GROUP   12
#define MARK            13
#define IP_TABLE        14
static int get_ipt_free(int what)
{
	FILE *pipe;

	/* 
	 * before the space on sed -es, I've used \\ because gcc
	 * read also as "bad escape sequence", this command simply
	 * return a list of used netlink group used on filter 
	 * table (and mark for mangle), after reading the pipe we search 
	 * the first netlink group free and use them
	 */
#define COMMAND_NETLINK "iptables -nL | grep ULOG | sed -es/.*nlgroup\\ //g | cut -b 1,2"
#define COMMAND_MARK "iptables -t mangle -nL | grep MARK | sed -es/.*MARK\\ set\\ // | printf \"%d\\n\" `tee`"
#define COMMAND_IPR2 "ip route show table all|grep table| sed -es/.*table\\ //g && cat /etc/iproute2/rt_tables | grep -v \\#"

#define MAXNUM 32
	unsigned char free_number[MAXNUM];
	int i =0x00;
	char *command;

	switch(what)
	{
		case NETLINK_GROUP:
			command =COMMAND_NETLINK;
			break;
		case MARK:
			command =COMMAND_MARK;
			break;
		case IP_TABLE:
			command =COMMAND_IPR2;
			break;
		default:
		/* make gcc happy and without warning :) */
			command =NULL;
			break;
	}

	if((pipe =popen(command, "r")) ==NULL)
		innova_p(FATAL, "unable to run [%s] for auto detection", command);

	memset((void *)free_number, 0x00, MAXNUM);

	while(!feof(pipe))
	{
		char str[4], *nl;
		int number;

		memset(str, 0x00, 4);

		fgets(str, 4, pipe);

		/* EOF */
		if(str[0] ==0x00)
			break;

		 /* invalid char (happend with iproute) */
		if(str[0] > 9 && str[0] < 0)
			continue;

		if((nl =strchr(str, '\n')) !=NULL)
			*nl =0x00;

		number =atoi(str);

		if(number < MAXNUM)
			free_number[number]++;

		/* else, continue, ... */
	}

	pclose(pipe);

	 /*
	 * is returned i +1 because array is since 0 to 31, but netlink
	 * valid group are from 1 to 32 ... marked follow a different counting,
	 * but, I'm traking used marker, I wish that none use more than 32 marked
	 * and want run innova. if this happens, mail me to vecna@s0ftpj.org :)
	 */
	for(i =0x00; i < MAXNUM; i++)
		if(!free_number[i])
			return (i +1);

	return -1;
}

static int open_datalink(char *interface)
{
	int fd;
	struct sockaddr_ll s_ll;
	struct ifreq desc;

	if((fd =socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) ==-1)
		innova_p(FATAL_ERRNO, "unable to open UDP socket");

	strncpy(desc.ifr_name, interface, IFNAMSIZ);

	if(ioctl(fd, SIOCGIFINDEX, &desc) ==-1)
		innova_p(FATAL_ERRNO, "unable to get index from %s:", interface);

	close(fd);

	if((fd =socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) ==-1)
		innova_p(FATAL_ERRNO, "unable to open PF_PACKET socket");

	s_ll.sll_family =PF_PACKET;
	s_ll.sll_protocol =htons(ETH_P_IP);
	s_ll.sll_ifindex = desc.ifr_ifindex;
	s_ll.sll_pkttype =PACKET_HOST;

	if(bind(fd, (struct sockaddr *)&s_ll, sizeof(s_ll)) ==-1)
		innova_p(FATAL_ERRNO, "unable to bind %s", interface);

	return fd;
}

void checked_execution(char *cmd)
{
	FILE *pipe;

	if((pipe =popen(cmd, "r")) ==NULL)
		innova_p(FATAL_ERRNO, "unable to pipe for command [%s]", cmd);

	pclose(pipe);
}

int
netlink_init(struct innova_struct *is, struct innova_options *io)
{
#define PORTOPTLEN	32
	char miscbuff[PORTOPTLEN];
	unsigned char proto;
	int h =1;

	/*
	 * opening of raw sock for resend packets
	 */
	proto =io->protocol->p_proto ? io->protocol->p_proto : IPPROTO_TCP;

	if((is->rawfd =socket(AF_INET, SOCK_RAW, proto)) ==-1)
		innova_p(FATAL_ERRNO, "opening raw socket");

	if(setsockopt(is->rawfd, IPPROTO_IP, IP_HDRINCL, &h, sizeof(h)) == -1)
		innova_p(FATAL_ERRNO, "unable to setsockopt");

	/* 
	 * generation of DROP rule for block ingoing remote packets,
	 * and for read them from datalink layer, on INPUT is impossibile 
	 * match with netfilter the source pid of software sending,
	 * for this reason innova change ttl to 255 and resend to him
	 * kernel via localhost, localhost shouldn't ave rp_filter 
	 * set, ttl 255 shouldn't happens on normal condition, default
	 * ttl si more lower and at each hop them is decremented.
	 */
	fill_portopt(io->retocol->p_proto, REMOTE, miscbuff, PORTOPTLEN, io->rem_serv, io->loc_serv);
	snprintf(ipt_cmd_str_remote, MAXCOMMANDLEN,
	"%s -A INPUT -s %s/%s %s -m ttl ! --ttl 255 -j DROP",
			IPTABLES_PATH,
			io->contact->h_name,
			io->network != NULL ? io->network->n_name : "255.255.255.255",
			io->remote_match_ext != NULL ? io->remote_match_ext : miscbuff
		);

	/*
	 * http://iptables-tutorial.frozentux.net/chunkyhtml/traversingoftables.html
	 * tnx to Braindead for this hints, I must analyze better this document, I've
	 * found them only when I'm solving loop/drop/etc problem between iptables 
	 * chains, should be impelmented a better solution also using /dev/netfilter_ipv4
	 * or better netlink/ulog rules.
	 *
	 * this is only innova first release :)
	 */

	/*
	 * getting of free netlink group, usable for identify locally 
	 * generated patckets, matching the rules host/service/protocol based
	 */
	if((is->local_nlg =get_ipt_free(NETLINK_GROUP)) ==-1)
		return innova_p(ERROR, "unable to get netlink group for local input\n");

	if((is->local_mark =get_ipt_free(MARK)) ==-1)
		return innova_p(ERROR, "unable to get netlink mark free for output\n");

	if((is->innova_table =get_ipt_free(IP_TABLE)) ==-1)
		return innova_p(ERROR, "unable to get innova free table from iproute2\n");
	
	/*
	 * OUTPUT mangle is called as first hook for outgoing packets, here I set with netfilter
	 * mark all packets matching innova rules but don't generated from innova
	 * (should possibile that innova resent packets with the some characteristic of the 
	 * packets described on the command line rules.
	 *
	 * here I set the mark on the packets but for undestound why, see below :)
	 */
	fill_portopt(io->protocol->p_proto, LOCAL, miscbuff, PORTOPTLEN, io->rem_serv, io->loc_serv);

	snprintf(ipt_cmd_str_local_mark, MAXCOMMANDLEN,
	"%s -A OUTPUT -t mangle -d %s/%s %s -m owner ! --pid-owner %d -j MARK --set-mark %d",
			IPTABLES_PATH,
			io->remote->h_name,
			io->network != NULL ? io->network->n_name : "255.255.255.255",
			io->local_match_ext != NULL ? io->local_match_ext : miscbuff,
			getpid(),
			is->local_mark
		);

	/* 
	 * rule for netlink ulog packets dest for target host, this packets doesn't appear marked 
	 * because marked packets is discarged, if they appear marked is not for innova marking
	 * but for previous, the mark should be analyzed in sturct innova_packet.mark 
	 */
	snprintf(ipt_cmd_str_local_ulog, MAXCOMMANDLEN,
	"%s -A OUTPUT -d %s/%s %s -m owner ! --pid-owner %d -j ULOG --ulog-nlgroup %d",
			IPTABLES_PATH,
			io->remote->h_name,
			io->network != NULL ? io->network->n_name : "255.255.255.255",
			io->local_match_ext != NULL ? io->local_match_ext : miscbuff,
			getpid(),
			is->local_nlg
		);

	/* 
	 * ULOG, how LOG target doesn't drop (or "stolen") the packets, simply duplicated
	 * it or report him ... but the packet generated from application must not be send,
	 * is required to be dropped. 
	 *
	 * use -j DROP for outgoing packets different from innova pid or without innova mark
	 * should seem to me an options, but have a problem: -j DROP if applied to locally
	 * generated packets return -EPERM error on the network syscall (connect(2) sendto(2)
	 * and other. this is a very big problem.
	 *
	 * with a sympatic hack I've used POSTROUTING hook on nat table (the last hook 
	 * analyzer for outgoing packets) who apply a MASQUERADE target, MASQUERADE recompute
	 * routing table.
	 *
	 * the routing table is changed below using iproute policy routing and netfilter
	 * mark as routing key, with this the packets matching the rule is routed as 
	 * specify on the policy routing (below is set to be localhost).
	 *
	 * with the packets resend locally because read via datalink layer, on the localhost
	 * ALL packets before innova mangling should be monitored, this is a very useful
	 * debug application.
	 */
	snprintf(ipt_cmd_str_local_masq, MAXCOMMANDLEN,
	"%s -A POSTROUTING -t nat -d %s/%s %s -m mark --mark %d -j MASQUERADE",
			IPTABLES_PATH,
			io->remote->h_name,
			io->network != NULL ? io->network->n_name : "255.255.255.255",
			io->local_match_ext != NULL ? io->local_match_ext : miscbuff,
			is->local_mark
		);

	/*
	 * rules MASQUERADE on the previous command is requider for recompute rule and 
	 * apply mark request rerouting, this cause who the packets outgoing matched doesn't
	 * going out with normale routing table, but is forwarded in DEFAULT_TRASH_INTERFACE,
	 * I'm using usually "lo", because is very useful have on localhost the traffic
	 * specific for innova (without innova modification) for debuggin pourpose
	 */
	snprintf(iproute_cmd_local_dup, MAXCOMMANDLEN,
	"%s rule add fwmark %d table %d && %s route add default via 127.0.0.1 dev %s table %d",
			IPROUTE_PATH,
			is->local_mark,
			is->innova_table,
			IPROUTE_PATH,
			DEFAULT_TRASH_INTERFACE,
			is->innova_table
		);

	checked_execution(ipt_cmd_str_local_mark);
	checked_execution(ipt_cmd_str_local_ulog);
	checked_execution(ipt_cmd_str_local_masq);
	checked_execution(iproute_cmd_local_dup);

	checked_execution(ipt_cmd_str_remote);

	/* this was about rule, now about netlink socket ... */
	if((is->local =ipulog_create_handle(ipulog_group2gmask(is->local_nlg))) ==NULL)
		innova_p(FATAL, "unable to create local netlink handle: %s", ipulog_strerror(errno));

	/* 
	 * for locally generated packets is used netlink filter, for remote 
	 * packets is runned read at datalink layer socket, iptables
	 * filter prevent kernels reply.
	 *
	 * open_datalink manage error self, also because if datalink opeing
	 * fail, no other choose should be take nor exit.
	 */
	is->lremotp =open_datalink(io->interface);
	return 0x00;
}

void netlink_cleanup()
{
	char *ip_action;
	const char *new_action ="del";

	IPTCMD_DELETE_RULES(ipt_cmd_str_local_mark);
	IPTCMD_DELETE_RULES(ipt_cmd_str_local_ulog);
	IPTCMD_DELETE_RULES(ipt_cmd_str_local_masq);
	IPTCMD_DELETE_RULES(ipt_cmd_str_remote);

	/*
	 * for delete ip rule add and ip route add, must became 
	 *            ip rule del and ip route del 
	 * :)
	 */
	while((ip_action =strstr(iproute_cmd_local_dup, "add")) !=NULL)
		memcpy(ip_action, new_action, strlen(new_action));

	system(iproute_cmd_local_dup);
}

/*
 * the various fill_*hdr are implemented as inline function on protocol.c
 */
inline int 
fill_pkt_pointer(struct innova_packet *pkt, unsigned char *pb)
{
	memset(pkt, 0x00, sizeof(struct innova_packet));

	pkt->ip =(struct iphdr *)pb;

	if(pkt->ip->version == 4)
	{
		int layer3_l, layer4_l;

		if((pkt->ip->ihl *4) > sizeof(struct iphdr))
			fill_ip_options(pkt, pb, &layer3_l);
		else
			layer3_l =(pkt->ip->ihl *4);

		pkt->length =ntohs(pkt->ip->tot_len);

		switch(pkt->ip->protocol)
		{
			case IPPROTO_TCP:
				fill_tcphdr(pkt, pb, &layer4_l);
				break;
			case IPPROTO_ICMP:
				fill_icmphdr(pkt, pb, &layer4_l);
				break;
			case IPPROTO_UDP:
				fill_udphdr(pkt, pb, &layer4_l);
				break;
			default:
				break;
		}

		pkt->data_len =(pkt->length -layer3_l -layer4_l);

		if(pkt->data_len > 0)
			pkt->data =(void *)&pb[layer3_l +layer4_l];

		return 0x00;
	}
	else
	{
		innova_p(INFO, "not IPv4 header or bad packet!");
		return 0xff;
	}
}

static inline unsigned char check_port_range(struct innova_packet *pkt, unsigned char who, struct servent *se)
{
	/* 
	 * *se should be se == NULL in plugin option, on case where
	 * the protocol is mangled and the reply isn't with source port
	 * equal to your preview dest port (how happens in TCP and UDP)
	 */
	if(se ==NULL)
		return 0x00;
	
	if(pkt->ip->protocol ==IPPROTO_TCP)
	{
		if(who ==LOCAL)
		{
			if(pkt->tcp->dest != se->s_port)
				return 0xff;
		}
		else
		{
			if(pkt->tcp->source != se->s_port)
				return 0xff;
		}
	}
	else if (pkt->ip->protocol ==IPPROTO_UDP)
	{
		if(who ==LOCAL)
		{
			if(pkt->udp->dest != se->s_port)
				return 0xff;
		}
		else
		{
			if(pkt->udp->source != se->s_port)
				return 0xff;
		}
	}

	return 0x00;
}

/* 
 * valid_innova_packet is checked for datalink readed packets,
 * for check if they are matching with the firewall rules and 
 * if them should be passed to the innova engine or is referret
 * to other connections, this function use "current" innova struct
 * global variable, declared and filled on innova.c
 */
static inline unsigned char valid_innova_packet(struct innova_packet *pkt)
{
	const unsigned char use_netmask = (current->o->network != NULL) ? 0xff: 0x00;
	const unsigned int hostmask = (current->o->network != NULL)
				       ? (*(unsigned *)current->o->remote->h_addr & current->o->network->n_net)
				       : 0x00;

	if(pkt->ip->ttl == 255)
		return 0x00;

	if(use_netmask)
	{
		if((pkt->ip->saddr & current->o->network->n_net) != hostmask)
			return 0x00;
	}
	else
	{
		if(pkt->ip->saddr != *(int *)current->o->remote->h_addr)
			return 0x00;
	}

	if(pkt->IO ==REMOTE)
	{
		if(current->o->retocol->p_proto)
			if(current->o->retocol->p_proto !=pkt->ip->protocol)
				return 0x00;
	}
	else
	{
		if(current->o->protocol->p_proto)
			if(current->o->protocol->p_proto !=pkt->ip->protocol)
				return 0x00;
	}

	if(check_port_range(pkt, REMOTE, current->o->rem_serv))
		return 0x00;

	if(check_port_range(pkt, LOCAL, current->o->loc_serv))
		return 0x00;

	/* else, the packet should be ok :) */
	return 0xff;
}

#if 0
static inline unsigned char check_pkt_trace(struct pkt_trace *p, struct innova_packet *pkt, unsigned int l4_v)
{
	if(p->id == pkt->ip->id && p->tot_len == pkt->ip->tot_len && p->l4_value == l4_v)
		return 0xff;

	return 0x00;
}

static inline unsigned char is_dup(int from, struct innova_packet *pkt)
{
	/* for the last packets */
	static struct pkt_trace last[2];
	/* for the packets before the last :) */
	unsigned int l4_val;
	
	switch(pkt->ip->protocol)
	{
		case IPPROTO_TCP:
			/* ISN */
			l4_val =pkt->tcp->seq;
			l4_val +=(pkt->tcp->syn + pkt->tcp->ack);
			break;
		case IPPROTO_UDP:
			/* source and dest port */
			l4_val =pkt->udp->dest;
			l4_val +=pkt->udp->source;
			break;
		case IPPROTO_ICMP:
			/* type and code */
			l4_val =pkt->icmp->type;
			l4_val +=pkt->icmp->code;
			break;
		default:
			l4_val =0x00;
			break;
	}

	if(check_pkt_trace(&last[from], pkt, l4_val))
		return 0xff;

	last[from].id =pkt->ip->id;
	last[from].tot_len =pkt->ip->tot_len;
	last[from].l4_value =l4_val;

	return 0x00;
}
#endif

inline struct innova_packet *get_ulog_packet(struct ipulog_handle *ipuh, unsigned char *pb)
{
	static struct innova_packet ret;
	ulog_packet_msg_t *um =NULL;
	int nbyte =0x00;

	/*
	 * ipulog_read should get more than one packet, this packets 
	 * are queued on libipulog, ipulog_get_packet get packet
	 * from this queue, on ulog program is used a while cicle
	 * for continue to get packet until return NULL, here
	 * I must return one packet each time, for this 
	 * reason the first action of get_ulog_packet is to check
	 * on the queue the presence of old packets and get 
	 * them, if this doesn't, try to read him from netlink
	 * descriptor and start to read them.
	 *
	 * after the first packet returned, on the wait_packet 
	 * routine is set static variable "packet_queued" for 
	 * sign if should exist queued packets, if this happens
	 * wait_packet doesn't enter on the recv/check loop but 
	 * call get_ulog_packet until them return NULL, the 
	 * NULL received on the second ipulog_get_packet below 
	 */
	if((um =ipulog_get_packet(ipuh, pb, nbyte)) ==NULL)
	{

		if((nbyte =ipulog_read(ipuh, pb, BUFSIZE, 1)) ==-1)
		{
			/*
			 * error different to EAGAIN is bad, but 
			 * EAGAIN mean that no now packet are present,
			 * NULL is returned
			 */
			innova_p(ERROR, "%s", ipulog_strerror(ipulog_errno));	
			return NULL;
		}

		if((um =ipulog_get_packet(ipuh, pb, nbyte)) ==NULL)
		{
			innova_p(ERROR, "%s", ipulog_strerror(ipulog_errno));
			return NULL;
		}
	}

	/*
	 * to avoid duplication, this should be fixed 
	 */

	fill_pkt_pointer(&ret, um->payload);

#if 0
	if(is_dup(LOCAL, &ret))
		return NULL;
#endif

	/* additional ulog message information */
	ret.usec =um->timestamp_usec;
	ret.sec =um->timestamp_sec;
	ret.mark =um->mark;
	ret.IO =LOCAL;

//	mtu_tuning(&ret);

	return (&ret);
}

inline struct innova_packet *get_datalink_packet(int fd, unsigned char *pb)
{
	static struct innova_packet ret;
	struct timeval tv;
	int nbyte;

	if((nbyte =read(fd, pb, BUFSIZE)) ==-1)
		innova_p(ERROR_ERRNO, "unable to read from datalink stream");

	/* should not happens, but ... */
	if(!nbyte)
		return NULL;

	fill_pkt_pointer(&ret, pb);

	ret.IO =REMOTE;

	if(!verify_ip_check(ret.ip))
		return NULL;

	if(!valid_innova_packet(&ret))
		return NULL;

#if 0
	/*
	 * to avoid duplication, this should be fixed 
	 */
	if(is_dup(REMOTE, &ret))
		return NULL;
#endif

	gettimeofday(&tv, NULL);

	ret.usec =tv.tv_usec;
	ret.sec =tv.tv_sec;
	ret.mark =0x00;

//	mtu_tuning(&ret);
	
	return (&ret);
}

/* 
 * dump_in_buffer is a function for dump the single portion of the
 * packets (ip header, ip options, tcp or udp ecc...) and 
 * compute checksum. for be faster dump_in_buffer work on 
 * a shared temporaney buffer used a swap memory, useful for
 * rapid function.
 *
 * this function is called from fix_raw_packet in reinject_packet function, 
 * for build buffer to send, return value is total length.
 *
 * read append_in_buffer and get_b_addr on utilf.c for know more
 */
inline int dump_in_buffer(struct innova_packet *pkt)
{
	int ret =0x00;

	/* memset to 0 shared buffer */
	append_in_buffer(NULL, 0x00);

	/* iphdr and transport layer are already fixed */
	ret =append_in_buffer(pkt->ip, sizeof(struct iphdr));

	if(pkt->ip_opt !=NULL)
		ret +=append_in_buffer(pkt->ip_opt, pkt->ip_opt_len);

	switch(pkt->ip->protocol)
		{
		case IPPROTO_TCP:
			ret +=append_in_buffer(pkt->tcp, sizeof(struct tcphdr));

			if(pkt->tcp_opt !=NULL)
				ret +=append_in_buffer(pkt->tcp_opt, pkt->tcp_opt_len);
			break;
		case IPPROTO_ICMP:
			ret +=append_in_buffer(pkt->icmp, sizeof(struct icmphdr));
			break;
		case IPPROTO_UDP:
			ret +=append_in_buffer(pkt->udp, sizeof(struct udphdr));
			break;
		default:
			break;
		}

	if(pkt->data !=NULL)
		ret +=append_in_buffer(pkt->data, pkt->data_len);

	return ret;
}

/*
 * for sign packets read from datalink layer coming from remote
 * host, after get_datalink_packet the data is resend via localhost
 * interface (routing optimization know destination address as
 * local address, but packets matching the rule match also the
 * DROP for incoming packets. the rule for drop incomimng packets
 * accept packets with ttl 255 (possibile only on local resend)
 */
static void inline fix_ttl_value(struct innova_packet *pkt)
{
	if(pkt->ip !=NULL)
		pkt->ip->ttl =255;
}

/*
 * compute checksum before dump packet, this is the function that
 * call dump_in_buffer, called by reinject_packet...
 */
static int inline fix_packet_raw(struct innova_packet *pkt)
	{
	int pktlen =0x00;

	fix_ip_checksum(pkt);

	switch(pkt->ip->protocol)
		{
		case IPPROTO_TCP:
			fix_tcp_checksum(pkt);
			break;
		case IPPROTO_ICMP:
			fix_icmp_checksum(pkt);
			break;
		case IPPROTO_UDP:
			fix_udp_checksum(pkt);
			break;
		default:
			innova_p(WARN,
				"unsupported protocol checksum %d",
				pkt->ip->protocol);
			break;
		}

	pktlen =dump_in_buffer(pkt);

	return pktlen;
	}

/*
 * function for send packet. all other information is on this file :)
 */
inline int reinject_packet(struct innova_packet *pkt, struct innova_struct *is)
{
	int len, ret =0x00;
	struct sockaddr_in sin;
	const int slen =sizeof(struct sockaddr_in);

	/*
	 * mangle function should have build a number of packets, not only one,
	 * innova_packet->next is used for this reason, allocation of addictional
	 * packets is managed by the plugin, here simple is follow che chain of
	 * packets using the "next" pointer until is NULL, return value is sum of
	 * all bytes sent, or if negative, error code returned frop innova_p
	 * (who return -errno)
	 */
	while(pkt !=NULL)
	{
		if(pkt->foo_packet_f !=NULL)
			pkt->foo_packet_f(is, pkt);

		if(pkt->IO ==REMOTE)
			fix_ttl_value(pkt);

		/* fix_packet_raw call dump_in_buffer that call append_in_buffer */
		len =fix_packet_raw(pkt);

		sin.sin_family =AF_INET;
		sin.sin_addr.s_addr =pkt->ip->daddr;

		if(sendto(is->rawfd, get_b_addr(), len, 0x00, (struct sockaddr *)&sin, slen) ==-1)
			return innova_p(ERROR_ERRNO, "unable to send via socket raw");
		else
			ret += len;

		pkt =pkt->next;
	}

	return ret;
}

/*
 * this is static declaration for MTU get/set, simply I didn't
 * build this MTU/MSS tracking during my debug activities,
 * I've not try to undestand if this is required or not, 
 * on the next release I should delete them, but is nice, if
 * you are interested, this is free software :)
 */
static struct ifreq imtudesc;

/*
 * this function is called any packet received and check if 
 * the MSS requested on tcp option header is lower than the 
 * actual MSS, if this happens is changed MTU related to the 
 * interface
 */
inline void mtu_tuning(struct innova_packet *pkt)
{
	int len, i;
	unsigned char opt;
	unsigned short tmp_mss;

	/* if are not present tcp options, should return immediatly */
	if(!pkt->tcp_opt_len)
		return;

	/* tcp options are formed with [KIND][LEN][OPTION_DATA] */
	for(i =0x00; i < (pkt->tcp_opt_len -2); i++)
	{
		opt =pkt->tcp_opt[i];

		/* skip len (1) and return */
		if(opt ==TCPOPT_NOP)
		{
			i++;
			continue;
		}
		
		if(opt ==TCPOPT_MAXSEG)
		{
			/* 
			 * check if length is ok or should be caused
			 * by invalide tcp option
			 */
			len =(int)pkt->tcp_opt[++i];

			if(len != 4)
				return;

			/* 
			 * copy MSS requested, add to iphdr and tcphdr size for
			 * compare effective MTU
			 */
			memcpy(&tmp_mss, &pkt->tcp_opt[++i], sizeof(short));
			tmp_mss =ntohs(tmp_mss)+ sizeof(struct iphdr)+ sizeof(struct tcphdr);

			if(tmp_mss < current->min_MSS)
				current->min_MSS =tmp_mss;
			else
				return;

			imtudesc.ifr_mtu =current->min_MSS;

			if(ioctl(current->ifd, SIOCSIFMTU, &imtudesc) ==-1)
				innova_p(FATAL_ERRNO, "unable to MTU tuning");
		}
		else
		{
			/* skip other options than MSS */
			len =(int)pkt->tcp_opt[++i];

			if(len >= 0)
				i +=len;
			else
				return;
		}
	}
}

/*
 * mtu_change change the MTU on our sending interface, simply them should
 * take two argument (mtu_delta and mtu_fixed) for manage them, mtu_delta
 * is a signed value for increment/decrement actual MTU, mtu_fixed is of 
 * set a fixed MTU
 */
unsigned int mtu_change(struct innova_struct *is, short mtu_delta, short mtu_fixed)
{
	if(mtu_delta && mtu_fixed)
		innova_p(FATAL, "bad use of mtu_change from plugin!");

	if(mtu_fixed)
		imtudesc.ifr_mtu =mtu_fixed;
	else
		imtudesc.ifr_mtu +=mtu_delta;

	if(ioctl(current->ifd, SIOCSIFMTU, &imtudesc) ==-1)
		innova_p(FATAL_ERRNO, "unable to MTU change");

	current->min_MSS =imtudesc.ifr_mtu;

	return imtudesc.ifr_mtu;
}

/*
 * this function is called after network initialization,
 * for get interfaces MTU
 */
unsigned int mtu_init(struct innova_struct *is, struct innova_options *io)
{
	unsigned int ret;

	if((is->ifd =socket(AF_INET, SOCK_DGRAM, IPPROTO_IP)) ==-1)
		innova_p(FATAL_ERRNO, "unable to open udp socket");

	strncpy(imtudesc.ifr_name, io->interface, IFNAMSIZ);

	if(ioctl(is->ifd, SIOCGIFMTU, &imtudesc) ==-1)
		innova_p(FATAL, "unable to get MTU from %s", imtudesc.ifr_name);

	ret =imtudesc.ifr_mtu;

	return ret;
}

/*
 * this function simply set the original MTU and 
 * close socket for manage interface
 */
void mtu_restore(struct innova_struct *is, unsigned int orig_mtu)
{
	imtudesc.ifr_mtu =orig_mtu;

	if(ioctl(is->ifd, SIOCSIFMTU, &imtudesc) ==-1)
		innova_p(FATAL, "unable to set MTU for %s", imtudesc.ifr_name);

	close(is->ifd);
}
